import 'package:org_parser/org_parser.dart';
import 'package:petitparser/petitparser.dart';
import 'package:petitparser/reflection.dart';
import 'package:test/test.dart';

import 'matchers.dart';

void main() {
  group('headline', () {
    final grammarDefinion = OrgGrammarDefinition();
    final parser = grammarDefinion.buildFrom(grammarDefinion.headline());
    test('parse a headline', () {
      final result = parser.parse('* Title');
      expect(
        result.value,
        ['* ', null, null, 'Title', null],
      );
    });
    test('parse almost-a-header', () {
      final result = parser.parse('**');
      expect(result is Failure, true);
    });
    test('parse just an empty header', () {
      final result = parser.parse('* ');
      expect(
        result.value,
        ['* ', null, null, null, null],
      );
    });
    test('parse a todo header', () {
      final result = parser.parse('* TODO Title');
      expect(
        result.value,
        ['* ', 'TODO', null, 'Title', null],
      );
    });
    test('parse a complex header', () {
      final result = parser.parse('** TODO [#A] Title foo bar :biz:baz:');
      expect(result.value, [
        '** ',
        'TODO',
        ['[#', 'A', ']'],
        'Title foo bar',
        [
          ' :',
          isSeparatedList(elements: [
            'biz',
            'baz',
          ], separators: [
            ':'
          ]),
          ':',
          null
        ]
      ]);
    });
  });
  group('structural grammar', () {
    final grammar = OrgGrammarDefinition().build();
    test('parse content', () {
      final result = grammar.parse('''foo
bar
''');
      expect(result.value, ['foo\nbar\n', []]);
    });
    test('parse an empty header before a regular header', () {
      final result = grammar.parse('''**${' '}
* foo''');
      expect(result.value, [
        null,
        [
          [
            ['** ', null, null, null, null],
            null
          ],
          [
            ['* ', null, null, 'foo', null],
            null
          ]
        ]
      ]);
    });
    test('parse an almost-header before content', () {
      final result = grammar.parse('''*
foo''');
      expect(result.value, ['*\nfoo', []]);
    });
    test('parse a section', () {
      final result = grammar.parse('''* Title
  Content1
  Content2''');
      expect(result.value, [
        null,
        [
          [
            ['* ', null, null, 'Title', null],
            '  Content1\n  Content2'
          ]
        ]
      ]);
    });
    test('valid headers', () {
      for (final valid in [
        '* ',
        '** DONE',
        '*** Some e-mail',
        '**** TODO [#A] COMMENT Title :tag:a2%:',
      ]) {
        expect(grammar.parse(valid) is Success, true);
      }
    });
  });

  group('content grammar parts', () {
    final grammarDefinition = OrgContentGrammarDefinition();
    Parser buildSpecific(Parser Function() start) {
      return grammarDefinition.buildFrom(start()).end();
    }

    test('paragraph', () {
      final parser = buildSpecific(grammarDefinition.paragraph);
      var result = parser.parse('''foo bar
biz baz''');
      expect(result.value, [
        '',
        ['foo bar\nbiz baz']
      ]);
      result = parser.parse('''go to [[http://example.com][example]] for *fun*,
maybe''');
      expect(result.value, [
        '',
        [
          'go to ',
          [
            '[',
            ['[', 'http://example.com', ']'],
            ['[', 'example', ']'],
            ']'
          ],
          ' for ',
          ['*', 'fun', '*'],
          ',\nmaybe'
        ]
      ]);
    });
    test('link', () {
      final parser = buildSpecific(grammarDefinition.link);
      var result = parser.parse('[[http://example.com][example]]');
      expect(result.value, [
        '[',
        ['[', 'http://example.com', ']'],
        ['[', 'example', ']'],
        ']'
      ]);
      result = parser.parse('[[*\\[wtf\\] what?][[lots][of][boxes]\u200b]]');
      expect(result.value, [
        '[',
        ['[', '*[wtf] what?', ']'],
        ['[', '[lots][of][boxes]', ']'],
        ']'
      ]);
      result = parser.parse('http://example.com');
      expect(result.value, 'http://example.com');
      result = parser.parse('https://example.com');
      expect(result.value, 'https://example.com');
      result = parser.parse('file:example.txt');
      expect(result.value, 'file:example.txt');
      result = parser.parse('foobar://example.com');
      expect(result is Failure, true);
    });
    test('markup', () {
      final parser = buildSpecific(grammarDefinition.markups);
      var result = parser.parse('''a/b
c/d''');
      expect(result is Failure, true, reason: 'bad pre/post chars');
      result = parser.parse('''a /b
c/d''');
      expect(result is Failure, true, reason: 'bad post char');
      result = parser.parse('''a/b
c/ d''');
      expect(result is Failure, true, reason: 'bad pre char');
      result = parser.parse('/a/');
      expect(result.value, ['/', 'a', '/']);
      result = parser.parse('/abc/');
      expect(result.value, ['/', 'abc', '/']);
      result = parser.parse('/a b/');
      expect(result.value, ['/', 'a b', '/']);
      result = parser.parse('//');
      expect(result is Failure, true, reason: 'body is required');
      result = parser.parse('~,~');
      expect(result.value, ['~', ',', '~']);
      result = parser.parse("~'~");
      expect(result.value, ['~', "'", '~']);
      result = parser.parse('=+LEVEL=3+boss-TODO​="DONE"=');
      expect(result.value, ['=', '+LEVEL=3+boss-TODO​="DONE"', '=']);
      result = parser.parse('''+foo
bar+''');
      expect(result.value, ['+', 'foo\nbar', '+']);
      result = parser.parse('''+foo

bar+''');
      expect(true, result is Failure);
    });
    test('macro reference', () {
      final parser = buildSpecific(grammarDefinition.macroReference);
      var result = parser.parse('{{{name(arg1, arg2)}}}');
      expect(result.value, ['{{{', 'name', '(arg1, arg2)', '}}}']);
      result = parser.parse('{{{foobar}}}');
      expect(result.value, ['{{{', 'foobar', '', '}}}']);
      result = parser.parse('{{{}}}');
      expect(result is Failure, true, reason: 'Body missing');
      result = parser.parse('{{{0abc}}}');
      expect(result is Failure, true, reason: 'Invalid key');
    });
    test('affiliated keyword', () {
      final parser = buildSpecific(grammarDefinition.affiliatedKeyword);
      var result = parser.parse('  #+blah');
      expect(result.value, ['  ', '#+blah', '']);
      result = parser.parse('''a   #+blah''');
      expect(result is Failure, true, reason: 'only leading space is allowed');
    });
    test('block', () {
      final parser = buildSpecific(grammarDefinition.block);
      var result = parser.parse('''#+begin_src sh
  echo 'foo'
  rm bar
#+end_src''');
      expect(result.value, [
        '',
        [
          [
            '#+begin_src',
            [' ', 'sh'],
            '\n'
          ],
          '  echo \'foo\'\n  rm bar\n',
          ['', '#+end_src']
        ],
        ''
      ]);
      result = parser.parse('''#+BEGIN_SRC sh
  echo 'foo'
  rm bar
#+EnD_sRC
''');
      expect(result.value, [
        '',
        [
          [
            '#+BEGIN_SRC',
            [' ', 'sh'],
            '\n'
          ],
          '  echo \'foo\'\n  rm bar\n',
          ['', '#+EnD_sRC']
        ],
        '\n'
      ]);
      result = parser.parse('''#+begin_src
  echo 'foo'
  rm bar
#+end_src
''');
      expect(result.value, [
        '',
        [
          ['#+begin_src', null, '\n'],
          '  echo \'foo\'\n  rm bar\n',
          ['', '#+end_src']
        ],
        '\n'
      ]);
    });
    test('greater block', () {
      final parser = buildSpecific(grammarDefinition.greaterBlock);
      var result = parser.parse('''#+begin_quote
  foo *bar*
#+end_quote''');
      expect(result.value, [
        '',
        [
          ['#+begin_quote', '\n'],
          [
            '  foo ',
            ['*', 'bar', '*'],
            '\n'
          ],
          ['', '#+end_quote']
        ],
        ''
      ]);
      result = parser.parse('''#+BEGIN_QUOTE
  foo /bar/
#+EnD_qUOtE
''');
      expect(result.value, [
        '',
        [
          ['#+BEGIN_QUOTE', '\n'],
          [
            '  foo ',
            ['/', 'bar', '/'],
            '\n'
          ],
          ['', '#+EnD_qUOtE']
        ],
        '\n'
      ]);
    });
    test('table', () {
      final parser = buildSpecific(grammarDefinition.table);
      final result = parser.parse('''  | foo | bar | baz |
  |-----+-----+-----|
  |   1 |   2 |   3 |
''');
      expect(result.value, [
        [
          [
            '  ',
            '|',
            [
              [
                ' ',
                ['foo'],
                ' |'
              ],
              [
                ' ',
                ['bar'],
                ' |'
              ],
              [
                ' ',
                ['baz'],
                ' |'
              ]
            ],
            '\n'
          ],
          ['  ', '|-----+-----+-----|\n'],
          [
            '  ',
            '|',
            [
              [
                '   ',
                ['1'],
                ' |'
              ],
              [
                '   ',
                ['2'],
                ' |'
              ],
              [
                '   ',
                ['3'],
                ' |'
              ]
            ],
            '\n'
          ]
        ],
        ''
      ]);
    });
    test('timestamps', () {
      final parser = buildSpecific(grammarDefinition.timestamp);
      var result = parser.parse('''<2020-03-12 Wed>''');
      expect(result.value, [
        '<',
        ['2020', '-', '03', '-', '12', 'Wed'],
        null,
        [],
        '>'
      ]);
      result = parser.parse('''<2020-03-12 Wed 8:34>''');
      expect(result.value, [
        '<',
        ['2020', '-', '03', '-', '12', 'Wed'],
        ['8', ':', '34'],
        [],
        '>'
      ]);
      result = parser.parse('''<2020-03-12 Wed 8:34 +1w>''');
      expect(result.value, [
        '<',
        ['2020', '-', '03', '-', '12', 'Wed'],
        ['8', ':', '34'],
        [
          ['+', '1', 'w']
        ],
        '>'
      ]);
      result = parser.parse('''<2020-03-12 Wed 8:34 +1w --2d>''');
      expect(result.value, [
        '<',
        ['2020', '-', '03', '-', '12', 'Wed'],
        ['8', ':', '34'],
        [
          ['+', '1', 'w'],
          ['--', '2', 'd']
        ],
        '>'
      ]);
      result = parser.parse('''[2020-03-12 Wed 18:34 .+1w --12d]''');
      expect(result.value, [
        '[',
        ['2020', '-', '03', '-', '12', 'Wed'],
        ['18', ':', '34'],
        [
          ['.+', '1', 'w'],
          ['--', '12', 'd']
        ],
        ']'
      ]);
      result = parser.parse('''[2020-03-12 Wed 18:34-19:35 .+1w --12d]''');
      expect(result.value, [
        '[',
        ['2020', '-', '03', '-', '12', 'Wed'],
        [
          ['18', ':', '34'],
          '-',
          ['19', ':', '35']
        ],
        [
          ['.+', '1', 'w'],
          ['--', '12', 'd']
        ],
        ']'
      ]);
      result = parser.parse(
          '''[2020-03-11 Wed 18:34 .+1w --12d]--[2020-03-12 Wed 18:34 .+1w --12d]''');
      expect(result.value, [
        [
          '[',
          ['2020', '-', '03', '-', '11', 'Wed'],
          ['18', ':', '34'],
          [
            ['.+', '1', 'w'],
            ['--', '12', 'd']
          ],
          ']'
        ],
        '--',
        [
          '[',
          ['2020', '-', '03', '-', '12', 'Wed'],
          ['18', ':', '34'],
          [
            ['.+', '1', 'w'],
            ['--', '12', 'd']
          ],
          ']'
        ]
      ]);
      result = parser.parse('''<%%(what (the (f)))>''');
      expect(result.value, [
        '<%%',
        [
          '(',
          [
            'what',
            [
              '(',
              [
                'the',
                [
                  '(',
                  ['f'],
                  ')'
                ]
              ],
              ')'
            ]
          ],
          ')'
        ],
        '>'
      ]);
      result = parser.parse('''<%%(what (the (f))>''');
      expect(result is Failure, true, reason: 'Invalid sexp');
      result = parser.parse('''[2020-03-11 Wed 18:34:56 .+1w --12d]''');
      expect(result is Failure, true, reason: 'Seconds not supported');
    });
    test('fixed-width area', () {
      final parser = buildSpecific(grammarDefinition.fixedWidthArea);
      var result = parser.parse('  : foo');
      expect(result.value, [
        [
          ['  ', ': ', 'foo']
        ],
        ''
      ]);
      result = parser.parse('''  : foo
  : bar''');
      expect(result.value, [
        [
          ['  ', ': ', 'foo\n'],
          ['  ', ': ', 'bar']
        ],
        ''
      ]);
    });
    test('list', () {
      final parser = buildSpecific(grammarDefinition.list);
      var result = parser.parse('- foo');
      expect(result.value, [
        [
          [
            '',
            [
              '- ',
              null,
              null,
              ['foo']
            ]
          ]
        ],
        ''
      ]);
      result = parser.parse('''- foo
  - bar''');
      expect(result.value, [
        [
          [
            '',
            [
              '- ',
              null,
              null,
              [
                'foo\n',
                [
                  [
                    [
                      '  ',
                      [
                        '- ',
                        null,
                        null,
                        ['bar']
                      ]
                    ]
                  ],
                  ''
                ]
              ]
            ]
          ]
        ],
        ''
      ]);
      result = parser.parse('''- foo

  bar''');
      expect(result.value, [
        [
          [
            '',
            [
              '- ',
              null,
              null,
              ['foo\n\n  bar']
            ]
          ]
        ],
        ''
      ]);
      result = parser.parse(
        '  - foo\n'
        ' \n'
        '    bar',
      );
      expect(result.value, [
        [
          [
            '  ',
            [
              '- ',
              null,
              null,
              ['foo\n \n    bar']
            ]
          ]
        ],
        ''
      ]);
      result = parser.parse('''30. [@30] foo
   - bar :: baz
     blah
   - [ ] *bazinga*''');
      expect(result.value, [
        [
          [
            '',
            [
              ['30', '.', ' '],
              ['[@', '30', ']'],
              null,
              [
                'foo\n',
                [
                  [
                    [
                      '   ',
                      [
                        '- ',
                        null,
                        [
                          ['bar'],
                          ' :: '
                        ],
                        ['baz\n     blah\n']
                      ]
                    ],
                    [
                      '   ',
                      [
                        '- ',
                        ['[', ' ', ']'],
                        null,
                        [
                          ['*', 'bazinga', '*']
                        ]
                      ]
                    ]
                  ],
                  ''
                ]
              ]
            ]
          ]
        ],
        ''
      ]);
      result = parser.parse('''- foo
  #+begin_src sh
    echo bar
  #+end_src''');
      expect(result.value, [
        [
          [
            '',
            [
              '- ',
              null,
              null,
              [
                'foo\n',
                [
                  '  ',
                  [
                    [
                      '#+begin_src',
                      [' ', 'sh'],
                      '\n'
                    ],
                    '    echo bar\n',
                    ['  ', '#+end_src']
                  ],
                  ''
                ]
              ]
            ]
          ]
        ],
        ''
      ]);
      result = parser.parse('- ~foo~ ::');
      expect(result.value, [
        [
          [
            '',
            [
              '- ',
              null,
              [
                [
                  ['~', 'foo', '~']
                ],
                ' ::'
              ],
              []
            ]
          ]
        ],
        ''
      ]);
    });
    test('planning line', () {
      final parser = buildSpecific(grammarDefinition.planningLine);
      final result = parser.parse(
          'CLOCK: [2021-02-16 Tue 21:40]--[2021-02-16 Tue 21:40] =>  0:00');
      expect(result.value, [
        '',
        [
          'CLOCK:',
          [
            ' ',
            [
              [
                '[',
                ['2021', '-', '02', '-', '16', 'Tue'],
                ['21', ':', '40'],
                [],
                ']'
              ],
              '--',
              [
                '[',
                ['2021', '-', '02', '-', '16', 'Tue'],
                ['21', ':', '40'],
                [],
                ']'
              ]
            ],
            ' =>  0:00'
          ]
        ],
        ''
      ]);
    });
    test('drawer', () {
      final parser = buildSpecific(grammarDefinition.drawer);
      var result = parser.parse(''':foo:
:end:''');
      expect(result.value, [
        '',
        [
          [':', 'foo', ':', '\n'],
          [],
          ['', ':end:']
        ],
        ''
      ]);
      result = parser.parse('''  :foo:
  :bar: baz
  :end:

''');
      expect(result.value, [
        '  ',
        [
          [':', 'foo', ':', '\n'],
          [
            [
              '  ',
              [':', 'bar', ':'],
              ' baz',
              '\n'
            ]
          ],
          ['  ', ':end:']
        ],
        '\n\n'
      ]);
      result = parser.parse(''':LOGBOOK:
CLOCK: [2021-01-23 Sat 09:30]--[2021-01-23 Sat 10:19] =>  0:49
:END:
''');
      expect(result.value, [
        '',
        [
          [':', 'LOGBOOK', ':', '\n'],
          [
            [
              '',
              [
                'CLOCK:',
                [
                  ' ',
                  [
                    [
                      '[',
                      ['2021', '-', '01', '-', '23', 'Sat'],
                      ['09', ':', '30'],
                      [],
                      ']'
                    ],
                    '--',
                    [
                      '[',
                      ['2021', '-', '01', '-', '23', 'Sat'],
                      ['10', ':', '19'],
                      [],
                      ']'
                    ]
                  ],
                  ' =>  0:49'
                ]
              ],
              '\n'
            ]
          ],
          ['', ':END:']
        ],
        '\n'
      ]);
      result = parser.parse(''':foo:
:bar:
:end:
:end:''');
      expect(result is Failure, true,
          reason: 'Nested drawer disallowed; the trailing ":end:" is '
              'a separate paragraph, which fails the drawer-specific parser');
    });
    test('property', () {
      final parser = buildSpecific(grammarDefinition.property);
      var result = parser.parse(':foo: bar');
      expect(result.value, [
        '',
        [':', 'foo', ':'],
        ' bar',
        ''
      ]);
      result = parser.parse(':foo:');
      expect(result is Failure, true, reason: 'Value required');
      result = parser.parse(':foo:blah');
      expect(result is Failure, true, reason: 'Delimiting space required');
      result = parser.parse(''':foo:
bar''');
      expect(result is Failure, true, reason: 'Value must be on same line');
    });
    test('footnote', () {
      final parser = buildSpecific(grammarDefinition.footnote);
      var result = parser.parse('[fn:1] foo *bar*');
      expect(result.value, [
        ['[fn:', '1', ']'],
        [
          ' foo ',
          ['*', 'bar', '*']
        ],
        ''
      ]);
      result = parser.parse('''[fn:1] foo *bar*
baz bazinga

''');
      expect(result.value, [
        ['[fn:', '1', ']'],
        [
          ' foo ',
          ['*', 'bar', '*'],
          '\nbaz bazinga\n\n'
        ],
        ''
      ]);
      result = parser.parse(' [fn:1] foo *bar*');
      expect(result is Failure, true, reason: 'Indent not allowed');
      result = parser.parse('[fn:1: blah] foo *bar*');
      expect(result is Failure, true, reason: 'Only simple references allowed');
    });
    test('footnote reference', () {
      final parser = buildSpecific(grammarDefinition.footnoteReference);
      var result = parser.parse('[fn:1]');
      expect(result.value, ['[fn:', '1', ']']);
      result = parser.parse('[fn:abc123]');
      expect(result.value, ['[fn:', 'abc123', ']']);
      result = parser.parse('[fn:abc123: who what why]');
      expect(result.value, [
        '[fn:',
        'abc123',
        ':',
        [' who what why'],
        ']'
      ]);
      result = parser.parse('[fn:abc123: who *what* why]');
      expect(result.value, [
        '[fn:',
        'abc123',
        ':',
        [
          ' who ',
          ['*', 'what', '*'],
          ' why'
        ],
        ']'
      ]);
    });
    test('entity', () {
      final parser = buildSpecific(grammarDefinition.entity);
      var result = parser.parse(r'\there4');
      expect(result.value, [r'\', 'there4', '']);
      result = parser.parse(r'\sup1');
      expect(result.value, [r'\', 'sup1', '']);
      result = parser.parse(r'\sup5');
      expect(result is Failure, true);
      result = parser.parse(r'\frac12');
      expect(result.value, [r'\', 'frac12', '']);
      result = parser.parse(r'\frac15');
      expect(result is Failure, true);
      result = parser.parse(r'\foobar');
      expect(result.value, [r'\', 'foobar', '']);
      result = parser.parse(r'\foobar2');
      expect(result is Failure, true);
      result = parser.parse(r'\foobar{}');
      expect(result.value, [r'\', 'foobar', '{}']);
    });
  });

  group('content grammar complete', () {
    final parser = OrgContentGrammarDefinition().build();
    test('paragraph', () {
      final result = parser.parse('''foo bar *biz*

  #+begin_quote
    blah
  #+end_quote
bazinga''');
      expect(result.value, [
        [
          '',
          [
            'foo bar ',
            ['*', 'biz', '*'],
            '\n\n'
          ]
        ],
        [
          '  ',
          [
            ['#+begin_quote', '\n'],
            ['    blah\n'],
            ['  ', '#+end_quote']
          ],
          '\n'
        ],
        [
          '',
          ['bazinga']
        ]
      ]);
    });
    test('link', () {
      var result = parser.parse('a http://example.com b');
      expect(result.value, [
        [
          '',
          ['a ', 'http://example.com', ' b']
        ]
      ]);
      result = parser.parse('a https://example.com b');
      expect(result.value, [
        [
          '',
          ['a ', 'https://example.com', ' b']
        ]
      ]);
      result = parser.parse('a [[foo][bar]] b');
      expect(result.value, [
        [
          '',
          [
            'a ',
            [
              '[',
              ['[', 'foo', ']'],
              ['[', 'bar', ']'],
              ']'
            ],
            ' b'
          ]
        ]
      ]);
      result = parser.parse('a [[foo::1][bar]] b');
      expect(result.value, [
        [
          '',
          [
            'a ',
            [
              '[',
              ['[', 'foo::1', ']'],
              ['[', 'bar', ']'],
              ']'
            ],
            ' b'
          ]
        ]
      ]);
    });
    test('affiliated keyword', () {
      var result = parser.parse('''#+blah
foo''');
      expect(result.value, [
        ['', '#+blah', '\n'],
        [
          '',
          ['foo']
        ]
      ]);
      result = parser.parse('''   #+blah
foo''');
      expect(result.value, [
        ['   ', '#+blah', '\n'],
        [
          '',
          ['foo']
        ]
      ]);
      // TODO(aaron): Figure out why this fails without the leading 'a'
      result = parser.parse('''a
#+blah
foo''');
      expect(result.value, [
        [
          '',
          ['a\n']
        ],
        ['', '#+blah', '\n'],
        [
          '',
          ['foo']
        ]
      ]);
    });
    test('list', () {
      final result = parser.parse('''- foo


  bar''');
      expect(result.value, [
        [
          [
            [
              '',
              [
                '- ',
                null,
                null,
                ['foo\n\n']
              ]
            ]
          ],
          '\n'
        ],
        [
          '  ',
          ['bar']
        ]
      ]);
    });
    test('drawer', () {
      final result = parser.parse(''':foo:
:bar: baz
:end:
#+bazinga: bozo''');
      expect(result.value, [
        [
          '',
          [
            [':', 'foo', ':', '\n'],
            [
              [
                '',
                [':', 'bar', ':'],
                ' baz',
                '\n'
              ]
            ],
            ['', ':end:']
          ],
          '\n'
        ],
        ['', '#+bazinga:', ' bozo']
      ]);
    });
    test('block', () {
      final result = parser.parse('''- foo
  #+begin_example
    bar


  #+end_example
  baz
''');
      expect(
        result.value,
        [
          [
            [
              [
                '',
                [
                  '- ',
                  null,
                  null,
                  [
                    'foo\n',
                    [
                      '  ',
                      [
                        ['#+begin_example', '\n'],
                        '    bar\n\n\n',
                        ['  ', '#+end_example']
                      ],
                      '\n'
                    ],
                    [
                      '  ',
                      ['baz\n']
                    ]
                  ]
                ]
              ]
            ],
            ''
          ]
        ],
        // This parse result is incompatible with the requirement that 3
        // linebreaks ends a list item, given the current implementation of
        // IndentedRegionParser
        //
        // TODO(aaron): Fix block parsing inside list item
        skip: true,
      );
    });
    group('arbitrary block', () {
      test('rich text content', () {
        final result = parser.parse('''#+begin_fofo
   bar ~baz~
#+end_fofo''');
        expect(result.value, [
          [
            '',
            [
              ['#+begin_', 'fofo', '\n'],
              [
                '   bar ',
                ['~', 'baz', '~'],
                '\n'
              ],
              ['', '#+end_fofo']
            ],
            ''
          ]
        ]);
      });
      test('https://github.com/amake/orgro/issues/79', () {
        final result = parser.parse('''#+BEGIN_RESULTS
Text
#+END_RESULTS''');
        expect(result.value, [
          [
            '',
            [
              ['#+BEGIN_', 'RESULTS', '\n'],
              ['Text\n'],
              ['', '#+END_RESULTS']
            ],
            ''
          ]
        ]);
      });
      test('mismatched capitalization', () {
        final result = parser.parse('''#+Begin_description
Text
#+end_DESCRIPTION''');
        expect(result.value, [
          [
            '',
            [
              ['#+Begin_', 'description', '\n'],
              ['Text\n'],
              ['', '#+end_DESCRIPTION']
            ],
            ''
          ]
        ]);
      });
    });
    test('LaTeX block', () {
      var result = parser.parse(r'''\begin{equation}
\nabla \cdot \mathbf{B} = 0
\end{equation}
''');
      expect(result.value, [
        [
          '',
          [
            [r'\begin{', 'equation', '}'],
            '\n\\nabla \\cdot \\mathbf{B} = 0\n',
            r'\end{equation}'
          ],
          '\n'
        ]
      ]);
      result = parser.parse(r'''\begin{equation}
\begin{cases}
   a &\text{if } b \\
   c &\text{if } d
\end{cases} + 1
\end{equation}
''');
      expect(result.value, [
        [
          '',
          [
            [r'\begin{', 'equation', '}'],
            '\n\\begin{cases}\n   a &\\text{if } b \\\\\n   c &\\text{if } d\n\\end{cases} + 1\n',
            r'\end{equation}'
          ],
          '\n'
        ]
      ]);
    });
    test('inline LaTeX', () {
      var result = parser.parse(r'foo $bar$ baz');
      expect(result.value, [
        [
          '',
          [
            'foo ',
            [r'$', 'bar', r'$'],
            ' baz'
          ]
        ]
      ]);
      result = parser.parse(r'from $i$ to $j$');
      expect(result.value, [
        [
          '',
          [
            'from ',
            [r'$', 'i', r'$'],
            ' to ',
            [r'$', 'j', r'$']
          ]
        ]
      ]);
      result = parser.parse(r'foo $$ a^2 + b^2 + c^2 $$ baz');
      expect(result.value, [
        [
          '',
          [
            'foo ',
            [r'$$', ' a^2 + b^2 + c^2 ', r'$$'],
            ' baz'
          ]
        ]
      ]);
      result = parser.parse(r'foo \(1/0\) baz');
      expect(result.value, [
        [
          '',
          [
            'foo ',
            [r'\(', '1/0', r'\)'],
            ' baz'
          ]
        ]
      ]);
      result = parser.parse(r'foo \[\infty\] baz');
      expect(result.value, [
        [
          '',
          [
            'foo ',
            [r'\[', r'\infty', r'\]'],
            ' baz'
          ]
        ]
      ]);
    });
    test('entity', () {
      var result = parser.parse(r'I think \there4 I am');
      expect(result.value, [
        [
          '',
          [
            'I think ',
            [r'\', 'there4', ''],
            ' I am'
          ]
        ]
      ]);
      result = parser.parse(r'''I think \there4
I am''');
      expect(result.value, [
        [
          '',
          [
            'I think ',
            [r'\', 'there4', ''],
            '\nI am'
          ]
        ]
      ]);
      result = parser.parse(r'I think \there4');
      expect(result.value, [
        [
          '',
          [
            'I think ',
            [r'\', 'there4', ''],
          ]
        ]
      ]);
      result = parser.parse(r'I think \there4{}');
      expect(result.value, [
        [
          '',
          [
            'I think ',
            [r'\', 'there4', '{}'],
          ]
        ]
      ]);
      result = parser.parse(r'I think \there4{}I am');
      expect(result.value, [
        [
          '',
          [
            'I think ',
            [r'\', 'there4', '{}'],
            'I am'
          ]
        ]
      ]);
    });
    test('detect common problems', () {
      expect(
        linter(parser, excludedTypes: {LinterType.info}),
        isEmpty,
        // TODO(aaron): There are two warnings about repeated choice, but
        // there's no hint as to where in the grammar they are.
        skip: true,
      );
    });
  });
  group('file link', () {
    final parser = OrgFileLinkGrammarDefinition().build();
    test('with scheme', () {
      var result = parser.parse('file:/home/dominik/images/jupiter.jpg');
      expect(result.value, ['file:', '/home/dominik/images/jupiter.jpg', null]);
      result = parser.parse('file:papers/last.pdf');
      expect(result.value, ['file:', 'papers/last.pdf', null]);
      result = parser.parse('file:/ssh:me@some.where:papers/last.pdf');
      expect(
        result.value,
        ['file:', '/ssh:me@some.where:papers/last.pdf', null],
      );
    });
    test('with extra', () {
      var result = parser.parse('file:sometextfile::123');
      expect(result.value, ['file:', 'sometextfile', '123']);
      result = parser.parse('file:projects.org::some words');
      expect(result.value, ['file:', 'projects.org', 'some words']);
      result = parser.parse('file:projects.org::*task title');
      expect(result.value, ['file:', 'projects.org', '*task title']);
      result = parser.parse('file:projects.org::#custom-id');
      expect(result.value, ['file:', 'projects.org', '#custom-id']);
      result = parser.parse('file:::*task title');
      expect(result.value, ['file:', '', '*task title']);
    });
    test('without scheme', () {
      var result = parser.parse('/home/dominik/images/jupiter.jpg');
      expect(result.value, ['', '/home/dominik/images/jupiter.jpg', null]);
      result = parser.parse('./papers/last.pdf');
      expect(result.value, ['', './papers/last.pdf', null]);
      result = parser.parse('./projects.org::#custom-id');
      expect(result.value, ['', './projects.org', '#custom-id']);
      result = parser.parse('../projects.org::#custom-id');
      expect(result.value, ['', '../projects.org', '#custom-id']);
    });
    test('non-files', () {
      var result = parser.parse('https://example.com');
      expect(result is Failure, true);
      result = parser.parse('mailto:me@example.com');
      expect(result is Failure, true);
    });
    test('detect common problems', () {
      expect(linter(parser), isEmpty);
    });
  });
}
